typedef union {
	uint8_t ub[4];
	uint16_t u16[2];
	uint32_t u32;
} uint32u_t;

typedef union {
	uint8_t ub[15];
	struct msc_cbw {  // Command Block Wrapper
		uint8_t  magic[4]; // "USBC"
		uint32_t tag;	
		volatile uint32u_t trans_len;
		uint8_t  flags;
		uint8_t  lun;
		uint8_t  cmd_len;
	} cbw;
	struct msc_csw { // Command Status Wrapper
		uint8_t  magic[4]; // "USBS"
		uint32_t tag;	
		uint32u_t trans_len;
		uint8_t  status;
	} csw;
} msc_cbw_t;

#define MSC_OWNER_PROG		1
#define MSC_OWNER_DRIVER	2

#ifdef EP_MS_OUT
extern void msc_init();
extern void msc_print_status();
extern void msc_poll();


extern void usbmsc_poll();

static inline int8_t usbmsc_can_getc(int8_t size) {
	UENUM = EP_MS_OUT;
	return (UEBCLX >= size);
}

static inline int8_t usbmsc_can_putc(int8_t size) {
	UENUM = EP_MS_IN;
	return (UEBCLX + size) <= EP_MS_IN_SIZE;
}

static inline unsigned char __usbmsc_getc(void) {
	unsigned char ret;
	UENUM = EP_MS_OUT;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbmsc_poll();
		UENUM = EP_MS_OUT;
	}
	ret = UEDATX;
	return ret;
}

static inline unsigned char __usbmsc_getc_nc(void) {
	UENUM = EP_MS_OUT;
	return UEDATX;
}


static inline void __usbmsc_putc(unsigned char data) {
	UENUM = EP_MS_IN;
	while (bit_is_clear(UEINTX,RWAL)) {
		usbmsc_poll();
		UENUM = EP_MS_IN;
	}
	UEDATX = data;
}

static inline void __usbmsc_putc_nc(unsigned char data) {
	UENUM = EP_MS_IN;
	UEDATX = data;
}

extern unsigned char usbmsc_getc(void);
extern void usbmsc_putc(unsigned char data);
extern void usbmsc_putc2(uint8_t ch);
extern void usbmsc_putc4(uint8_t ch);
extern void usbmsc_putc2_same(uint8_t ch);

extern uint8_t msc_lock_stat;

static inline uint8_t msc_is_locked(uint8_t owner) {
	if (owner == MSC_OWNER_PROG) {
		return (msc_lock_stat == MSC_OWNER_DRIVER);
	}
	return (msc_lock_stat == MSC_OWNER_PROG);
}

static inline void msc_lock(uint8_t owner) {
	if (owner == MSC_OWNER_PROG) {
		while (msc_is_locked(owner)) {
			msc_poll();
		}
	}
	msc_lock_stat = owner;
}

static inline void msc_unlock() {
	msc_lock_stat = 0;
}

/**/
extern uint8_t msc_sense_key[];
extern uint8_t msc_sense_asc[];

#define MSC_UNIT_NOT_READY              0
#define MSC_UNIT_CHANGE_TO_READY        1
#define MSC_UNIT_READY                  2

extern uint8_t msc_unit_state[];

/**/

extern msc_cbw_t msc_cbw;
extern volatile uint32u_t msc_trans_offset;

#define MSC_TRANS_CONT  0
#define MSC_TRANS_DONE  1
extern uint8_t (* msc_trans_proc)();
extern uint8_t msc_trans_skip();

extern int8_t msc_devattach(void (*func)(uint8_t,uint8_t));

#else // EP_MS_OUT
static inline void msc_lock(uint8_t owner) {}
static inline void msc_unlock() {}
#endif

#define MSC_SECTOR_SHIFT    9 // 512 B
//#define MSC_SECTOR_SHIFT 12 // 4KB
#define MSC_SECTOR_SIZE    (1<<MSC_SECTOR_SHIFT)


// Commands
#define SC_TEST_UNIT_READY                   (0x00)
#define SC_REQUEST_SENSE                     (0x03)
#define SC_INQUIRY                           (0x12)
#define SC_MODE_SENSE_6                      (0x1A)
#define SC_MODE_SENSE_10                     (0x5A)
#define SC_PREVENT_ALLOW_MEDIUM_REMOVAL      (0x1E)
#define SC_READ_CAPACITY                     (0x25)
#define SC_READ_FORMAT_CAPACITY              (0x23)
#define SC_READ_10                           (0x28)
#define SC_WRITE_10                          (0x2A)
#define SC_VERIFY_10                         (0x2F)
#define SC_START_STOP_UNIT                   (0x1B)

#if 0
#define SC_FORMAT_UNIT                       (0x04)
#define SC_READ_6                            (0x08)
#define SC_MODE_SELECT_10                    (0x55)
#define SC_RESERVE_10                        (0x56)
#define SC_MODE_SELECT_6                     (0x15)
#define SC_RELEASE_10                        (0x57)
#define SC_RECEIVE_DIAGNOSTICS               (0x1C)
#define SC_SEND_DIAGNOSTIC                   (0x1D)
#define SC_READ_CD_ROM_CAPACITY              (0x25)
#define SC_SEEK                              (0x2B)
#define SC_WRITE_VERIFY_10                   (0x2E)
#define SC_SYNCHRONIZE_CACHE                 (0x35)
#define SC_READ_TOC                          (0x43)
#endif

// Sense Key Code (msc_sense_key)
#define SC_SENSE_KEY_NO_SENSE                        (0x00)
#define SC_SENSE_KEY_NOT_READY                       (0x02)
#define SC_SENSE_KEY_ILLEGAL_REQUEST                 (0x05)
#define SC_SENSE_KEY_UNIT_ATTENTION                  (0x06)
#define SC_SENSE_KEY_MEDIUM_ERROR                    (0x03)
#if 0
#define SC_SENSE_KEY_RECOVERED_ERROR                 (0x01)
#define SC_SENSE_KEY_HARDWARE_ERROR                  (0x04)
#define SC_SENSE_KEY_DATA_PROTECT                    (0x07)
#define SC_SENSE_KEY_BLANK_CHECK                     (0x08)
#define SC_SENSE_KEY_VENDOR_SPECIFIC                 (0x09)
#define SC_SENSE_KEY_COPY_ABORTED                    (0x0A)
#define SC_SENSE_KEY_ABORTED_COMMAND                 (0x0B)
#define SC_SENSE_KEY_VOLUME_OVERFLOW                 (0x0D)
#define SC_SENSE_KEY_MISCOMPARE                      (0x0E)
#endif

// ASC code assignments (msc_sense_asc)
#define SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION       (0x00)
#define SC_ASC_LOGICAL_UNIT_NOT_READY                (0x04)
#define SC_ASC_INVALID_COMMAND_OPERATION_CODE        (0x20)
#define SC_ASC_INVALID_FIELD_IN_CDB                  (0x24)
#define SC_ASC_WRITE_PROTECTED                       (0x27)
#define SC_ASC_NOT_READY_TO_READY_CHANGE             (0x28)
#define SC_ASC_MEDIUM_NOT_PRESENT                    (0x3A)
#if 0
#define SC_ASC_FORMAT_ERROR                          (0x31)
#endif

// MODE SENSE and REQUEST SENSE DEFINITIONS                                  */

#define SC_MSK_DBD                                   (0x08)
#define SC_MSK_PAGE_CONTROL                          (0xC0)
#define SC_MSK_PAGE_CODE                             (0x3F)


// Mode page codes for direct-access devices (list in SBC specification)
#define SC_PAGE_CODE_ALL                             (0x3F) // SPC
#define SC_PAGE_CODE_CACHING_MODE_PAGE		     (0x08) // SBC
#define SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY       (0x01) // SBC
#define SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS        (0x1C) // SPC


